#!/usr/bin/env python
#
# SpamBayes launcher script (used by ThunderBayes)
# 
# This module is an extension to the spambayes project, which is Copyright
# 2002-3 The Python Software Foundation and is covered by the Python Software
# Foundation license.

import os
import re
import shutil
import socket
import sys

MAIL_ID_REGEX = re.compile(r"^X-Spambayes-MailId: ([0-9\-]+)$", re.M)

def main():
	thisdir = os.path.dirname(__file__)
	sys.path.append(thisdir)

	# setup data directory
	assert len(sys.argv) > 1, ("sb_launch takes at least one argument: "
		"the directory for spambayes data")
	datadir = sys.argv.pop(-1)
	if not os.path.exists(datadir):
		template = os.path.join(thisdir, "data_template")
		shutil.copytree(template, datadir)
	assert os.path.isdir(datadir), "bad data directory: %s" % datadir
	os.chdir(datadir)

	# import here to prevent options from loading before cwd is datadir
	# (any import from the spambayes package causes options to be loaded)
	import sb_server
	sb_server.ProxyUserInterface = addREST(sb_server.ProxyUserInterface) # HACK monkey patch
	sb_server.run()

def addREST(ProxyUserInterface):
	import spambayes.tokenizer as tokenizer
	from spambayes.Options import options, optionsPathname
	
	def log(msg, *args):
		if args:
			print msg % args
		else:
			print msg

	class ThunderBayesUserInterface(ProxyUserInterface):
	
		def onThunderbayes(self):
			self.writeOKHeaders('text/plain')
			self.write("ThunderBayes REST API")
		
		def onThunderbayesTrain(self, which, text):
			"""Train on text"""
			which = which.lower()
			isSpam = (which == 'spam')
	
			# Convert platform-specific line endings into unix-style.
			content = text.replace('\r\n', '\n').replace('\r', '\n')
			
			if not self.trainWithMessageKey(content, isSpam):
				# Append the message to a file, to make it easier to rebuild
				# the database later.   This is a temporary implementation -
				# it should keep a Corpus of trained messages.
				if isSpam:
					f = open("_pop3proxyspam.mbox", "a")
				else:
					f = open("_pop3proxyham.mbox", "a")
				try:
					self.flush()
					tokens = tokenizer.tokenize(content)
					self.classifier.learn(tokens, isSpam)
					f.write("From pop3proxy@spambayes.org Sat Jan 31 00:00:00 2000\n")
					f.write(content)
					f.write("\n\n")
				finally:
					f.close()
			self._doSave()
			self.writeOKHeaders('text/plain')
			self.write("done.")

		def trainWithMessageKey(self, message, isSpam):
			"""Train using embedded X-Spambayes-MailId header
			
			returns True if training succeeded, otherwise False.
			"""
			from spambayes.ProxyUI import state
			headers = message.split("\n\n", 1)[0]
			match = MAIL_ID_REGEX.search(headers)
			if not match:
				return False # mailId not found
			key = match.groups()[0]
			if state.spamCorpus.get(key) is not None:
				sourceCorpus = state.spamCorpus
			elif state.unknownCorpus.get(key) is not None:
				sourceCorpus = state.unknownCorpus
			elif state.hamCorpus.get(key) is not None:
				sourceCorpus = state.hamCorpus
			else:
				return False # unknown mailId
			if isSpam:
				targetCorpus = state.spamCorpus
			else:
				targetCorpus = state.hamCorpus
			targetCorpus.takeMessage(key, sourceCorpus, fromCache=True)
			return True

		def getThunderbayesPOP3ProxyInfo(self):
			ports = options["pop3proxy", "listen_ports"]
			servers = options["pop3proxy", "remote_servers"]
			assert len(ports) == len(servers), "pop3proxy_servers and pop3proxy_ports are different lengths!"
			return ports, servers

		def onThunderbayesPop3proxyinfo(self, proxyport=None, remote_server=None):
			proxyinfo = zip(*self.getThunderbayesPOP3ProxyInfo())
			if proxyport is None and remote_server is not None:
				# get proxy port associated with remote_server
				for proxyport, server in proxyinfo:
					if remote_server == server:
						self.writeOKHeaders('text/plain')
						self.write(proxyport)
						return
				error = remote_server + " has no proxy"
			elif proxyport is not None and remote_server is None:
				# get remote server associated with proxyport
				try:
					portinfo = dict(proxyinfo)[proxyport]
				except KeyError:
					error = "unknown proxy port: %r" % proxyport
				else:
					self.writeOKHeaders('text/plain')
					self.write(portinfo)
					return
			else:
				error = "invalid query string"
			self.thunderbayesError(500, error)
		
		def onThunderbayesUpdatepop3proxyinfo(self, remote_servers):
			"""Update POP3 proxy settings
			
			info - a comma-delimited list of remote servers to proxy
			       Example: mail.test.com:110,mail.xyz.com:110
			returns a comma-delimited list of proxy ports and remote servers
			       Example: 53100:mail.test.com:110,53101:mail.xyz.com:110
			"""
			ports, servers = self.getThunderbayesPOP3ProxyInfo()
			portmap = dict(zip(ports, servers))
			srvrmap = dict(zip(servers, ports))
			newservers = {}
			output = []
			for server in remote_servers.split(","):
				if not (server and server.strip()):
					continue
				try:
					proxyport = srvrmap[server]
				except KeyError:
					proxyport = 53100
					while proxyport < 53200:
						if str(proxyport) not in portmap:
							try:
								try:
									sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
									sock.bind(('', proxyport))
								finally:
									sock.close()
							except:
								pass
							else:
								break
						proxyport += 1
					proxyport = str(proxyport)
				newservers[server] = proxyport
				output.append("%s:%s" % (proxyport, server))
			if srvrmap != newservers:
				servers, ports = [], []
				for server, port in newservers.items():
					servers.append(server)
					ports.append(port)
				options.set("pop3proxy", "listen_ports", tuple(ports))
				options.set("pop3proxy", "remote_servers", tuple(servers))
				options.update_file(optionsPathname)
				self.reReadOptions()
			self.writeOKHeaders('text/plain')
			self.write(",".join(output))

		def thunderbayesError(self, code, message):
			headers = [
				"HTTP/1.0 %d Error" % code,
				"Connection: close",
				"Content-Type: text/plain",
				"",
				"",
			]
			self.write('\r\n'.join(headers) + message)
	
	return ThunderBayesUserInterface

if __name__ == "__main__":
	main()
